import logging

import numpy as np

logger = logging.getLogger()


def vec_normalize(vectors, /, *, out=None, dtype=None):
    """
    Normalize an array of vectors.

    Parameters
    ----------
    vectors : array_like, [..., 3]
        array of vectors
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [..., 3]
        array of normalized vectors.
    """
    vectors = np.asarray(vectors, dtype=np.float64)
    if out is None:
        out = np.empty_like(vectors, dtype=dtype)

    lengths_shape = vectors.shape[:-1] + (1,)
    lengths = np.linalg.norm(vectors, axis=-1).reshape(lengths_shape)
    return np.divide(vectors, lengths, out=out)


def vec_homogeneous(vectors, /, *, w=1, out=None, dtype=None):
    """
    Append homogeneous coordinates to vectors.

    Parameters
    ----------
    vectors : array_like, [..., 3]
        array of vectors
    w : number, optional, default is 1
        the value for the homogeneous dimensionality.
        this affects the result of translation transforms. use 0 (vectors)
        if the translation component should not be applied, 1 (positions)
        otherwise.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [..., 4]
        The list of vectors with appended homogeneous value.
    """
    vectors = np.asarray(vectors)
    shape = list(vectors.shape)
    shape[-1] += 1
    if out is None:
        out = np.empty_like(vectors, shape=shape, dtype=dtype)
    out[..., -1] = w
    out[..., :-1] = vectors
    return out


def vec_transform(vectors, matrix, /, *, w=1, out=None, dtype=None):
    """
    Apply a transformation matrix to a vector.

    Parameters
    ----------
    vectors : ndarray, [3]
        Array of vectors
    matrix : ndarray, [4, 4]
        Transformation matrix
    w : ndarray, [1], optional
        The value of the scale component of the homogeneous coordinate. This
        affects the result of translation transforms. use 0 (vectors) if the
        translation component should not be applied, 1 (positions) otherwise.
    out : ndarray, optional
        A location into which the result is stored. If provided, it must have a
        shape that the inputs broadcast to. If not provided or None, a
        freshly-allocated array is returned. A tuple must have length equal to
        the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [3]
        transformed vectors
    """

    vectors = np.asarray(vectors, dtype=float)
    matrix = np.asarray(matrix, dtype=float)

    if out is None:
        out_shape = np.broadcast_shapes(vectors.shape[:-1], matrix.shape[:-2])
        out = np.empty((*out_shape, 3), dtype=dtype)

    vectors = vec_homogeneous(vectors, w=w)
    result = matrix @ vectors[..., None]
    result /= result[..., -1, :][..., None, :]
    out[:] = result[..., :-1, 0]

    return out


def vec_unproject(vector, matrix, /, *, depth=0, out=None, dtype=None):
    """
    Un-project a vector from 2D space to 3D space.

    Find a ``vectorB`` in 3D euclidean space such that the projection
    ``matrix @ vectorB`` yields the provided vector (in 2D euclidean
    space). Since the solution to the above is a 1D subspace of 3D space (a
    line), ``depth`` is used to select a single vector within.

    Parameters
    ----------
    vector : ndarray, [2]
        The vector to be un-projected.
    matrix: ndarray, [4, 4]
        The camera's intrinsic matrix.
    depth : number, optional
        The distance of the unprojected vector from the camera.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    projected_vector : ndarray, [3]
        The unprojected vector in 3D space

    Notes
    -----
    The source frame of this operation is the XY-plane of the camera's NDC frame
    and the target frame is the camera's local frame.
    """

    vector = np.asarray(vector, dtype=float)
    matrix = np.asarray(matrix, dtype=float)
    depth = np.asarray(depth, dtype=float)

    result_shape = np.broadcast_shapes(
        vector.shape[:-1], matrix.shape[:-2], depth.shape
    )

    if out is None:
        out = np.empty((*result_shape, 3), dtype=dtype)

    try:
        inverse_projection = np.linalg.inv(matrix)
    except np.linalg.LinAlgError as err:
        raise ValueError("The provided matrix is not invertible.") from err

    vector_hom = np.empty((*result_shape, 4), dtype=dtype)
    vector_hom[..., 2] = depth
    vector_hom[..., [0, 1]] = vector
    vector_hom[..., 3] = 1

    out_hom = vector_hom @ inverse_projection.T
    scale = out_hom[..., -1][..., None]
    out[:] = (out_hom / scale)[..., :-1]

    return out


def vec_transform_quat(vector, quaternion, /, *, out=None, dtype=None):
    """Rotate a vector using a quaternion.

    Parameters
    ----------
    vector : ndarray, [3]
        The vector to be rotated.
    quaternion : ndarray, [4]
        The quaternion to apply (in xyzw format).
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    rotated_vector : ndarray, [3]
        The rotated vector.

    """

    vector = np.asarray(vector, dtype=float)
    quaternion = np.asarray(quaternion, dtype=float)

    if out is None:
        out = np.zeros_like(vector, dtype=dtype)

    # based on https://gamedev.stackexchange.com/a/50545
    # (more readable than my attempt at doing the same)

    quat_vector = quaternion[..., :-1]
    quat_scalar = quaternion[..., -1]

    out += 2 * np.sum(quat_vector * vector, axis=-1, keepdims=True) * quat_vector
    out += (
        quat_scalar**2 - np.sum(quat_vector * quat_vector, axis=-1, keepdims=True)
    ) * vector
    out += 2 * quat_scalar * np.cross(quat_vector, vector)

    return out


def vec_spherical_to_euclidean(spherical, /, *, out=None, dtype=None):
    """Convert spherical -> euclidean coordinates.

    Parameters
    ----------
    spherical : ndarray, [3]
        A vector in spherical coordinates (r, phi, theta). Phi and theta are
        measured in radians.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    euclidean : ndarray, [3]
        A vector in euclidean coordinates.

    Notes
    -----
    This implementation follows pygfx's coordinate conventions. This means that
    the positive y-axis is the zenith reference and the positive z-axis is the
    azimuth reference. Angles are measured counter-clockwise.

    """

    spherical = np.asarray(spherical, dtype=float)

    if out is None:
        out = np.empty_like(spherical, dtype=dtype)

    r, theta, phi = np.split(spherical, 3, axis=-1)
    out[..., 0] = r * np.sin(phi) * np.sin(theta)
    out[..., 1] = r * np.cos(phi)
    out[..., 2] = r * np.sin(phi) * np.cos(theta)

    return out


def vec_dist(vector_a, vector_b, /, *, out=None, dtype=None):
    """The distance between two vectors

    Parameters
    ----------
    vector_a : ndarray, [3]
        The first vector.
    vector_b : ndarray, [3]
        The second vector.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    distance : ndarray
        The distance between both vectors.

    """

    vector_a = np.asarray(vector_a, dtype=float)
    vector_b = np.asarray(vector_b, dtype=float)

    shape = vector_a.shape[:-1]
    if out is None:
        out = np.linalg.norm(vector_a - vector_b, axis=-1).astype(dtype)
    elif len(shape) >= 0:
        out[:] = np.linalg.norm(vector_a - vector_b, axis=-1)
    else:
        raise ValueError("Can't use `out` with scalar output.")

    return out


def vec_angle(vector_a, vector_b, /, *, out=None, dtype=None):
    """The angle between two vectors

    Parameters
    ----------
    vector_a : ndarray, [3]
        The first vector.
    vector_b : ndarray, [3]
        The second vector.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    angle : float
        The angle between both vectors.

    """

    vector_a = np.asarray(vector_a, dtype=float)
    vector_b = np.asarray(vector_b, dtype=float)

    shape = vector_a.shape[:-1]

    # Cannot broadcast np.dot(), so just write it out
    dot_prod = sum(
        [
            vector_a[..., 0] * vector_b[..., 0],
            vector_a[..., 1] * vector_b[..., 1],
            vector_a[..., 2] * vector_b[..., 2],
        ]
    )
    the_cos = (
        dot_prod / np.linalg.norm(vector_a, axis=-1) / np.linalg.norm(vector_b, axis=-1)
    )

    if out is None:
        out = np.arccos(the_cos).astype(dtype)
    elif len(shape) >= 0:
        out[:] = np.arccos(the_cos)
    else:
        raise ValueError("Can't use `out` with scalar output.")

    return out


def mat_decompose_translation(homogeneous_matrix, /, *, out=None, dtype=None):
    """Position component of a homogeneous matrix.

    Parameters
    ----------
    homogeneous_matrix : ndarray, [4, 4]
        The matrix of which the position/translation component will be
        extracted.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    position : ndarray, [3]
        The position/translation component.

    """

    homogeneous_matrix = np.asarray(homogeneous_matrix, dtype=float)

    if out is None:
        out = np.empty((*homogeneous_matrix.shape[:-2], 3), dtype=dtype)

    out[:] = homogeneous_matrix[..., :-1, -1]

    return out


def vec_euclidean_to_spherical(euclidean, /, *, out=None, dtype=None):
    """Convert euclidean -> spherical coordinates

    Parameters
    ----------
    euclidean : ndarray, [3]
        A vector in euclidean coordinates.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    spherical : ndarray, [3]
        A vector in spherical coordinates (r, phi, theta).

    """

    euclidean = np.asarray(euclidean, dtype=float)

    if out is None:
        out = np.zeros_like(euclidean, dtype=dtype)
    else:
        out[:] = 0

    out[..., 0] = np.sqrt(np.sum(euclidean**2, axis=-1))

    # flags to handle all cases
    needs_flip = np.sign(euclidean[..., 0]) < 0
    len_xz = np.sum(euclidean[..., [0, 2]] ** 2, axis=-1)
    xz_nonzero = ~np.all(len_xz == 0, axis=-1)
    r_nonzero = ~np.all(out[..., [0]] == 0, axis=-1)

    # chooses phi = 0 if vector runs along y-axis
    out[..., 1] = np.divide(euclidean[..., 2], np.sqrt(len_xz), where=xz_nonzero)
    out[..., 1] = np.arccos(out[..., 1], where=xz_nonzero)
    out[..., 1] = np.where(needs_flip, np.abs(out[..., 1] - np.pi), out[..., 1])

    # chooses theta = 0 at the origin (0, 0, 0)
    out[..., 2] = np.divide(euclidean[..., 1], out[..., 0], where=r_nonzero)
    out[..., 2] = np.arccos(out[..., 2], where=r_nonzero)
    out[..., 2] = np.where(needs_flip, 2 * np.pi - out[..., 2], out[..., 2])

    return out


def vec_spherical_safe(vector, /, *, out=None, dtype=None):
    """Normalize sperhical coordinates.

    Normalizes a vector of spherical coordinates to restrict phi to [0, pi) and
    theta to [0, 2pi).

    Parameters
    ----------
    vector : ndarray, [3]
        A vector in spherical coordinates.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    normalized_vector : ndarray, [3]
        A vector in spherical coordinates with restricted angle values.

    """

    vector = np.asarray(vector, dtype=float)

    if out is None:
        out = np.zeros_like(vector, dtype=dtype)

    is_flipped = vector[..., 1] % (2 * np.pi) >= np.pi
    out[..., 2] = np.where(is_flipped, -vector[..., 2], vector[..., 2])

    out[..., 0] = vector[..., 0]
    out[..., 1] = vector[..., 1] % np.pi
    out[..., 2] = vector[..., 1] % (2 * np.pi)

    out[..., 1] = np.where(out[..., 1] == np.pi, 0, out[..., 1])
    out[..., 2] = np.where(out[..., 2] == 2 * np.pi, 0, out[..., 2])

    return out


def quat_to_euler(quaternion, /, *, order="xyz", out=None, dtype=None):
    """Convert quaternions to Euler angles with specified rotation order.

    Parameters
    ----------
    quaternion : ndarray, [4]
        The quaternion to convert (in xyzw format).
    order : str, optional
        The rotation order as a string. Can include 'X', 'Y', 'Z' for intrinsic
        rotation (uppercase) or 'x', 'y', 'z' for extrinsic rotation (lowercase).
        Default is "xyz".
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    out : ndarray, [3]
        The Euler angles in the specified order.

    References
    ----------
    This implementation is based on the method described in the following paper:
    Bernardes E, Viollet S (2022) Quaternion to Euler angles conversion: A direct, general and computationally efficient method.
    PLoS ONE 17(11): e0276302. https://doi.org/10.1371/journal.pone.0276302
    """
    quaternion = np.asarray(quaternion, dtype=float)
    quat = np.atleast_2d(quaternion)
    num_rotations = quat.shape[0]

    if out is None:
        out = np.empty((*quaternion.shape[:-1], 3), dtype=dtype)

    extrinsic = order.islower()
    order = order.lower()
    if not extrinsic:
        order = order[::-1]

    basis_index = {"x": 0, "y": 1, "z": 2}
    i = basis_index[order[0]]
    j = basis_index[order[1]]
    k = basis_index[order[2]]

    is_proper = i == k

    if is_proper:
        k = 3 - i - j  # get third axis

    # Step 0
    # Check if permutation is even (+1) or odd (-1)
    sign = int((i - j) * (j - k) * (k - i) / 2)

    eps = 1e-7
    for ind in range(num_rotations):
        if num_rotations == 1 and out.ndim == 1:
            _angles = out
        else:
            _angles = out[ind, :]

        if is_proper:
            a = quat[ind, 3]
            b = quat[ind, i]
            c = quat[ind, j]
            d = quat[ind, k] * sign
        else:
            a = quat[ind, 3] - quat[ind, j]
            b = quat[ind, i] + quat[ind, k] * sign
            c = quat[ind, j] + quat[ind, 3]
            d = quat[ind, k] * sign - quat[ind, i]

        n2 = a**2 + b**2 + c**2 + d**2

        # Step 3
        # Compute second angle...
        # _angles[1] = 2*np.arccos(np.sqrt((a**2 + b**2) / n2))
        _angles[1] = np.arccos(2 * (a**2 + b**2) / n2 - 1)

        # ... and check if it is equal to 0 or pi, causing a singularity
        safe1 = np.abs(_angles[1]) >= eps
        safe2 = np.abs(_angles[1] - np.pi) >= eps
        safe = safe1 and safe2

        # Step 4
        # compute first and third angles, according to case
        if safe:
            half_sum = np.arctan2(b, a)  # == (alpha+gamma)/2
            half_diff = np.arctan2(-d, c)  # == (alpha-gamma)/2

            _angles[0] = half_sum + half_diff
            _angles[2] = half_sum - half_diff

        else:
            # _angles[0] = 0

            if not extrinsic:
                # For intrinsic, set first angle to zero so that after reversal we
                # ensure that third angle is zero
                # 6a
                if not safe:
                    _angles[0] = 0
                if not safe1:
                    half_sum = np.arctan2(b, a)
                    _angles[2] = 2 * half_sum
                # 6c
                if not safe2:
                    half_diff = np.arctan2(-d, c)
                    _angles[2] = -2 * half_diff
            else:
                # For extrinsic, set third angle to zero
                # 6b
                if not safe:
                    _angles[2] = 0
                if not safe1:
                    half_sum = np.arctan2(b, a)
                    _angles[0] = 2 * half_sum
                # 6c
                if not safe2:
                    half_diff = np.arctan2(-d, c)
                    _angles[0] = 2 * half_diff

        for i_ in range(3):
            if _angles[i_] < -np.pi:
                _angles[i_] += 2 * np.pi
            elif _angles[i_] > np.pi:
                _angles[i_] -= 2 * np.pi

        # for Tait-Bryan angles
        if not is_proper:
            _angles[2] *= sign
            _angles[1] -= np.pi / 2

        if not extrinsic:
            # reversal
            _angles[0], _angles[2] = _angles[2], _angles[0]

        # Step 8
        if not safe:
            logger.warning(
                "Gimbal lock detected. Setting third angle to zero "
                "since it is not possible to uniquely determine "
                "all angles."
            )

    return out


_warned_about_eucledian = False


def _warn_about_eucledian():
    global _warned_about_eucledian
    if not _warned_about_eucledian:
        _warned_about_eucledian = True
        m = "methods vec_spherical_to_euclidian() and vec_euclidian_to_spherical()"
        m += " have been renamed, because its eucledEan, not eucledIan."
        logger.warning("pylinalg deprecation warning: " + m)


def vec_euclidian_to_spherical(euclidean, /, *, out=None, dtype=None):
    _warn_about_eucledian()
    return vec_euclidean_to_spherical(euclidean, out=out, dtype=dtype)


def vec_spherical_to_euclidian(spherical, /, *, out=None, dtype=None):
    _warn_about_eucledian()
    return vec_spherical_to_euclidean(spherical, out=out, dtype=dtype)


__all__ = [
    name for name in globals() if name.startswith(("vec_", "mat_", "quat_", "aabb_"))
]



import numpy as np
from numpy.lib.stride_tricks import as_strided


def mat_combine(matrices, /, *, out=None, dtype=None):
    """
    Combine a list of affine matrices by multiplying them.

    Note that by matrix multiplication rules, the output matrix will applied the
    given transformations in reverse order. For example, passing a scaling,
    rotation and translation matrix (in that order), will lead to a combined
    transformation matrix that applies translation first, then rotation and finally
    scaling.

    Parameters
    ----------
    matrices : list of ndarray, [4, 4]
        List of affine matrices to combine.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4, 4]
        Combined transformation matrix.
    """

    matrices = [np.asarray(matrix) for matrix in matrices]
    result_shape = np.broadcast_shapes(*[matrix.shape for matrix in matrices])

    if out is None:
        out = np.empty(result_shape, dtype=dtype)

    out[:] = matrices[0]
    for matrix in matrices[1:]:
        np.matmul(out, matrix, out=out)

    return out


def mat_from_translation(vector, /, *, out=None, dtype=None):
    """
    Make a translationmatrix given a translation vector.

    Parameters
    ----------
    vector : number or ndarray, [3]
        translation vector
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4, 4]
        Translation matrix.
    """
    vector = np.asarray(vector)
    result_shape = (*vector.shape[:-1], 4, 4)

    if out is None:
        out = np.empty(result_shape, dtype=dtype)

    # view into the diagonal of the result
    n_matrices = np.prod(result_shape[:-2], dtype=int)
    itemsize = out.itemsize
    diagonal = as_strided(
        out, shape=(n_matrices, 4), strides=(16 * itemsize, 5 * itemsize)
    )

    out[:] = 0
    diagonal[:] = 1
    out[..., :-1, -1] = vector

    return out


def mat_from_scale(factors, /, *, out=None, dtype=None):
    """
    Make a scaling matrix given scaling factors per axis, or a
    single uniform scaling factor.

    Parameters
    ----------
    factor : number or ndarray, [3]
        scaling factor(s)
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4, 4]
        Scaling matrix.
    """
    factors = np.asarray(factors, dtype=dtype)

    matrix = np.identity(4, dtype=dtype)
    matrix[np.diag_indices(3)] = factors

    if out is not None:
        out[:] = matrix
        return out

    return matrix


def mat_from_euler(angles, /, *, order="xyz", out=None, dtype=None):
    """
    Make a matrix given euler angles (in radians) per axis.

    Parameters
    ----------
    angles : ndarray, [3]
        The euler angles.
    order : string, optional
        The rotation order as a string. Can include 'X', 'Y', 'Z' for intrinsic
        rotation (uppercase) or 'x', 'y', 'z' for extrinsic rotation (lowercase).
        Default is "xyz".
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4, 4]
        Rotation matrix.


    Notes
    -----
    If you are familiar with TreeJS note that this function uses ``order`` to
    denote both the order in which rotations are applied *and* the order in
    which angles are provided in ``angles``. I.e.,
    ``mat_from_euler([np.pi, np.pi, 0], order="zyx")``
    will first rotate 180° ccw (counter-clockwise) around the z-axis, then 180°
    ccw around the y-axis, and finally 0° around the x axis.

    """
    fill_out_first = out is not None
    angles = np.atleast_1d(np.asarray(angles))
    extrinsic = order.islower()
    order = order.upper()
    axis_lookup = {"X": 0, "Y": 1, "Z": 2}
    if extrinsic:
        angles = reversed(angles)
        order = reversed(order)
    for angle, axis in zip(angles, order):
        axis_idx = axis_lookup[axis]
        affine_matrix = np.identity(4, dtype=dtype)

        if axis_idx == 0:
            affine_matrix[1, 1] = np.cos(angle)
            affine_matrix[1, 2] = -np.sin(angle)
            affine_matrix[2, 1] = np.sin(angle)
            affine_matrix[2, 2] = np.cos(angle)
        elif axis_idx == 1:
            affine_matrix[0, 0] = np.cos(angle)
            affine_matrix[0, 2] = np.sin(angle)
            affine_matrix[2, 0] = -np.sin(angle)
            affine_matrix[2, 2] = np.cos(angle)
        elif axis_idx == 2:
            affine_matrix[0, 0] = np.cos(angle)
            affine_matrix[0, 1] = -np.sin(angle)
            affine_matrix[1, 0] = np.sin(angle)
            affine_matrix[1, 1] = np.cos(angle)

        if fill_out_first:
            out[:] = affine_matrix
            fill_out_first = False
        elif out is None:
            out = affine_matrix
        else:
            out @= affine_matrix
    return out


def mat_from_axis_angle(axis, angle, /, *, out=None, dtype=None):
    """
    Make a rotation matrix given a rotation axis and an angle (in radians).

    Parameters
    ----------
    axis : ndarray, [3]
        The rotation axis.
    angle : number
        The angle (in radians) to rotate about the axis.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4, 4]
        Rotation matrix.
    """
    axis = np.asarray(axis)

    if out is None:
        out = np.identity(4, dtype=dtype)
    else:
        out[:] = np.identity(4)

    eye = out[:3, :3]
    rotation = np.cos(angle) * eye
    # the second component here is the "cross product matrix" of axis
    rotation += np.sin(angle) * np.cross(axis, eye * -1)
    rotation += (1 - np.cos(angle)) * (np.outer(axis, axis))
    out[:3, :3] = rotation

    return out


def quat_from_mat(matrix, /, *, out=None, dtype=None):
    """
    Make a quaternion given a rotation matrix.

    Parameters
    ----------
    matrix : ndarray, [3]
        The rotation matrix.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4]
        Quaternion.
    """
    m = matrix[:3, :3]
    t = np.trace(m)

    if t > 0:
        s = 0.5 / np.sqrt(t + 1)
        x = (m[2, 1] - m[1, 2]) * s
        y = (m[0, 2] - m[2, 0]) * s
        z = (m[1, 0] - m[0, 1]) * s
        w = 0.25 / s

    elif m[0, 0] > m[1, 1] and m[0, 0] > m[2, 2]:
        s = 2 * np.sqrt(1 + m[0, 0] - m[1, 1] - m[2, 2])
        x = 0.25 * s
        y = (m[0, 1] + m[1, 0]) / s
        z = (m[0, 2] + m[2, 0]) / s
        w = (m[2, 1] - m[1, 2]) / s

    elif m[1, 1] > m[2, 2]:
        s = 2 * np.sqrt(1 + m[1, 1] - m[0, 0] - m[2, 2])
        x = (m[0, 1] + m[1, 0]) / s
        y = 0.25 * s
        z = (m[1, 2] + m[2, 1]) / s
        w = (m[0, 2] - m[2, 0]) / s

    else:
        s = 2 * np.sqrt(1 + m[2, 2] - m[0, 0] - m[1, 1])
        x = (m[0, 2] + m[2, 0]) / s
        y = (m[1, 2] + m[2, 1]) / s
        z = 0.25 * s
        w = (m[1, 0] - m[0, 1]) / s

    if out is None:
        out = np.empty((4,), dtype=dtype)
    out[:] = np.array([x, y, z, w])
    return out


def mat_compose(translation, rotation, scaling, /, *, out=None, dtype=None):
    """
    Compose a transformation matrix given a translation vector, a
    quaternion and a scaling vector.

    Parameters
    ----------
    translation : number or ndarray, [3]
        translation vector
    rotation : ndarray, [4]
        quaternion
    scaling : number or ndarray, [3]
        scaling factor(s)
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4, 4]
        Transformation matrix
    """
    if out is None:
        out = np.empty((4, 4), dtype=dtype)

    x, y, z, w = rotation
    x2 = x + x
    y2 = y + y
    z2 = z + z
    xx = x * x2
    xy = x * y2
    xz = x * z2
    yy = y * y2
    yz = y * z2
    zz = z * z2
    wx = w * x2
    wy = w * y2
    wz = w * z2

    scaling = np.asarray(scaling)
    if scaling.size == 1:
        scaling = np.broadcast_to(scaling, (3,))
    sx, sy, sz = scaling

    out[0, 0] = (1 - (yy + zz)) * sx
    out[1, 0] = (xy + wz) * sx
    out[2, 0] = (xz - wy) * sx
    out[3, 0:3] = 0

    out[0, 1] = (xy - wz) * sy
    out[1, 1] = (1 - (xx + zz)) * sy
    out[2, 1] = (yz + wx) * sy

    out[0, 2] = (xz + wy) * sz
    out[1, 2] = (yz - wx) * sz
    out[2, 2] = (1 - (xx + yy)) * sz

    out[0:3, 3] = translation
    out[3, 3] = 1

    return out


def mat_decompose(matrix, /, *, scaling_signs=None, dtype=None, out=None):
    """
    Decompose a transformation matrix into a translation vector, a
    quaternion and a scaling vector.

    Parameters
    ----------
    matrix : ndarray, [4, 4]
        transformation matrix
    scaling_signs : ndarray, [3], optional
        scaling factor signs. If you wish to preserve the original scaling
        factors through a compose-decompose roundtrip, you should
        provide the original scaling factors here, or alternatively just
        the signs.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    translation : ndarray, [3]
        translation vector
    rotation : ndarray, [4]
        quaternion
    scaling : ndarray, [3]
        scaling factors
    """
    matrix = np.asarray(matrix)

    if out is not None:
        translation = out[0]
    else:
        translation = np.empty((3,), dtype=dtype)
    translation[:] = matrix[:-1, -1]

    flip = 1 if np.linalg.det(matrix) >= 0 else -1
    if scaling_signs is not None:
        # if the user provides the scaling signs, always use them
        scaling_signs = np.sign(scaling_signs)
        if np.prod(scaling_signs) != flip:
            raise ValueError(
                "Number of negative signs is inconsistent with the determinant"
            )
    else:
        # if not, detect if a flip is needed to reconstruct the transform
        # and apply it to the first axis arbitrarily
        scaling_signs = np.array([flip, 1, 1])

    if out is not None:
        scaling = out[2]
    else:
        scaling = np.empty((3,), dtype=dtype)
    scaling[:] = np.linalg.norm(matrix[:-1, :-1], axis=0)
    scaling *= scaling_signs

    rotation = out[1] if out is not None else None

    rotation_matrix = matrix[:-1, :-1].copy().astype(float)
    mask = scaling != 0
    rotation_matrix[:, mask] /= scaling[mask][None, :]
    rotation_matrix[:, ~mask] = 0.0
    rotation = quat_from_mat(rotation_matrix, out=rotation, dtype=dtype)

    return translation, rotation, scaling


def mat_perspective(
    left, right, top, bottom, near, far, /, *, depth_range=(-1, 1), out=None, dtype=None
):
    """
    Create a perspective projection matrix.

    Parameters
    ----------
    left : number
        distance between the left frustum plane and the origin
    right : number
        distance between the right frustum plane and the origin
    top : number
        distance between the top frustum plane and the origin
    bottom : number
        distance between the bottom frustum plane and the origin
    near : number
        distance between the near frustum plane and the origin
    far : number
        distance between the far frustum plane and the origin
    depth_range : Tuple[float, float]
        The interval along the z-axis in NDC that shall correspond to the region
        inside the viewing frustum.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    matrix : ndarray, [4, 4]
        perspective projection matrix
    """
    if out is None:
        out = np.zeros((4, 4), dtype=dtype)
    else:
        out[:] = 0.0

    x = 2 * near / (right - left)
    y = 2 * near / (top - bottom)

    near_d = near * depth_range[0]
    far_d = far * depth_range[1]
    depth_diff = depth_range[1] - depth_range[0]

    a = (right + left) / (right - left)
    b = (top + bottom) / (top - bottom)
    c = -(far_d - near_d) / (far - near)
    d = -(far * near * depth_diff) / (far - near)

    out[0, 0] = x
    out[0, 2] = a
    out[1, 1] = y
    out[1, 2] = b
    out[2, 2] = c
    out[2, 3] = d
    out[3, 2] = -1

    return out


def mat_orthographic(
    left, right, top, bottom, near, far, /, *, depth_range=(-1, 1), out=None, dtype=None
):
    """Create an orthographic projection matrix.

    The result projects points from local space into NDC (normalized device
    coordinates). Elements inside the viewing frustum defind by left, right,
    top, bottom, near, far, are projected into the unit cube centered at the
    origin (default) or a cuboid (custom `depth_range`). The frustum is centered
    around the local frame's origin.

    Parameters
    ----------
    left : ndarray, [1]
        Distance between the left frustum plane and the origin
    right : ndarray, [1]
        Distance between the right frustum plane and the origin
    top : ndarray, [1]
        Distance between the top frustum plane and the origin
    bottom : ndarray, [1]
        Distance between the bottom frustum plane and the origin
    near : ndarray, [1]
        Distance between the near frustum plane and the origin
    far : ndarray, [1]
        Distance between the far frustum plane and the origin
    depth_range : ndarray, [2]
        The interval along the z-axis in NDC that shall correspond to the region
        inside the viewing frustum.
    out : ndarray, optional
        A location into which the result is stored. If provided, it must have a
        shape that the inputs broadcast to. If not provided or None, a
        freshly-allocated array is returned. A tuple must have length equal to
        the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    matrix : ndarray, [4, 4]
        orthographic projection matrix

    Notes
    -----
    The parameters to this function are given in a left-handed frame that is
    obtained by mirroring source's Z-axis at the origin. In other words, if the
    returned matrix represents a camera's projection matrix then this function's
    parameters are given in a frame that is like the camera's local frame except
    that it's Z-axis is inverted. This means that positive values for `near` and
    `far` refer to a negative Z values in camera local.

    """

    left = np.asarray(left, dtype=float)
    right = np.asarray(right, dtype=float)
    top = np.asarray(top, dtype=float)
    bottom = np.asarray(bottom, dtype=float)
    far = np.asarray(far, dtype=float)
    near = np.asarray(near, dtype=float)
    depth_range = np.asarray(depth_range, dtype=float)

    if out is None:
        batch_shape = np.broadcast_shapes(
            left.shape[:-1],
            right.shape[:-1],
            top.shape[:-1],
            bottom.shape[:-1],
            far.shape[:-1],
            near.shape[:-1],
            depth_range.shape[:-1],
        )
        out = np.zeros((*batch_shape, 4, 4), dtype=dtype)
    else:
        out[:] = 0

    # desired cuboid dimensions
    out[..., 0, 0] = 2
    out[..., 1, 1] = 2
    out[..., 2, 2] = -np.diff(depth_range, axis=-1)
    out[..., 3, 3] = 1

    # translation to cuboid origin
    out[..., 0, 3] = -(right + left)
    out[..., 1, 3] = -(top + bottom)
    out[..., 2, 3] = far * depth_range[..., 0] - near * depth_range[..., 1]

    # frustum-based scaling
    out[..., 0, :] /= right - left
    out[..., 1, :] /= top - bottom
    out[..., 2, :] /= far - near

    return out


def mat_look_at(eye, target, up_reference, /, *, out=None, dtype=None):
    """
    Rotation that aligns two vectors.

    Given an entity at position `eye` looking at position `target`, this
    function computes a rotation matrix that makes the local frame "look at" the
    same direction, i.e., the matrix will rotate the local frame's z-axes
    (forward) to point in direction ``target - eye``.

    This rotation matrix is not unique (yet), as the above doesn't specify the
    desired rotation around the new z-axis. The rotation around this axis is
    controlled by ``up_reference``, which indicates the direction of the y-axis
    (up) of a reference frame of choice expressed in local coordinates. The
    rotation around the new z-axis will then align `up_reference`, the new
    y-axis, and the new z-axis in the same plane.

    In many cases, a natural choice for ``up_reference`` is the world frame's
    y-axis, i.e., ``up_reference`` would be the world's y-axis expressed in
    local coordinates. This can be thought of as "gravity pulling on the
    rotation" (opposite direction of world frame's up) and will create a result
    with a level attitude.


    Parameters
    ----------
    eye : ndarray, [3]
        A vector indicating the direction that should be aligned.
    target : ndarray, [3]
        A vector indicating the direction to align on.
    up : ndarray, [3]
        The direction of the camera's up axis.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.


    Returns
    -------
    rotation_matrix : ndarray, [4, 4]
        A homogeneous matrix describing the rotation.

    Notes
    -----
    If the new z-axis (``target - eye``) aligns with the chosen ``up_reference``
    then we can't compute the angle of rotation around the new z-axis. In this
    case, we will default to a rotation of 0, which may result in surprising
    behavior for some use-cases. It is the user's responsibility to ensure that
    these two directions don't align.

    """

    eye = np.asarray(eye, dtype=float)
    target = np.asarray(target, dtype=float)
    up_reference = np.asarray(up_reference, dtype=float)

    new_z = target - eye
    up_reference = up_reference / np.linalg.norm(up_reference, axis=-1)

    result_shape = np.broadcast_shapes(eye.shape, target.shape, up_reference.shape)
    if out is None:
        out = np.zeros((*result_shape[:-1], 4, 4), dtype=dtype)
    else:
        out[:] = 0

    # Note: The below is equivalent to np.fill_diagonal(out, 1, axes=(-2, -1)),
    # i.e., treat the last two axes as a matrix and fill its diagonal with 1.
    # Currently numpy doesn't support axes on fill_diagonal, so we do it
    # ourselves to support input batches and mimic the `np.linalg` API.
    n_matrices = np.prod(result_shape[:-1], dtype=int)
    itemsize = out.itemsize
    view = as_strided(out, shape=(n_matrices, 4), strides=(16 * itemsize, 5 * itemsize))
    view[:] = 1

    out[..., :-1, 2] = new_z / np.linalg.norm(new_z, axis=-1)
    out[..., :-1, 0] = np.cross(
        up_reference, out[..., :-1, 2], axisa=-1, axisb=-1, axisc=-1
    )
    out[..., :-1, 1] = np.cross(
        out[..., :-1, 2], out[..., :-1, 0], axisa=-1, axisb=-1, axisc=-1
    )
    out /= np.linalg.norm(out, axis=-2)[..., None, :]

    return out


__all__ = [
    name for name in globals() if name.startswith(("vec_", "mat_", "quat_", "aabb_"))
]



"""
pylinalg

Linear algebra utilities for Python.
"""

from importlib.metadata import version

from .matrix import *
from .misc import *
from .quaternion import *
from .vector import *

__version__ = version("pylinalg")
version_info = tuple(map(int, __version__.split(".")))

del version


__all__ = [
    name for name in globals() if name.startswith(("vec_", "mat_", "quat_", "aabb_"))
]



import numpy as np


def aabb_to_sphere(aabb, /, *, out=None, dtype=None):
    """A sphere that envelops an Axis-Aligned Bounding Box.

    Parameters
    ----------
    aabb : ndarray, [2, 3]
        The axis-aligned bounding box.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    sphere : ndarray, [4]
        A sphere (x, y, z, radius).

    """

    aabb = np.asarray(aabb, dtype=float)

    if out is None:
        out = np.empty((*aabb.shape[:-2], 4), dtype=dtype)

    out[..., :3] = np.sum(aabb, axis=-2) / 2
    out[..., 3] = np.linalg.norm(np.diff(aabb, axis=-2), axis=-1) / 2

    return out


def aabb_transform(aabb, matrix, /, *, out=None, dtype=None):
    """Apply an affine transformation to an axis-aligned bounding box.

    Parameters
    ----------
    aabb : ndarray, [2, 3]
        The axis-aligned bounding box.
    homogeneous_matrix : [4, 4]
        The homogeneous transformation to apply.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    aabb : ndarray, [2, 3]
        The transformed axis-aligned bounding box.

    Notes
    -----
    This function preserves the alignment axes of the bounding box. This means
    the returned bounding box has the same alignment axes as the input bounding
    box, but contains the transformed object. In other words, the box will grow
    or shrink depending on how the contained object is transformed, but its
    alignment axis stay the same.

    """

    aabb = np.asarray(aabb, dtype=float)
    matrix = np.asarray(matrix, dtype=float).transpose((-1, -2))

    if out is None:
        out = np.empty_like(aabb, dtype=dtype)

    corners = np.full(
        aabb.shape[:-2] + (8, 4),
        # Fill value of 1 is used for homogeneous coordinates.
        fill_value=1.0,
        dtype=float,
    )
    # x
    corners[..., 0::2, 0] = aabb[..., 0, 0]
    corners[..., 1::2, 0] = aabb[..., 1, 0]

    # y
    corners[..., 0::4, 1] = aabb[..., 0, 1]
    corners[..., 1::4, 1] = aabb[..., 0, 1]
    corners[..., 2::4, 1] = aabb[..., 1, 1]
    corners[..., 3::4, 1] = aabb[..., 1, 1]

    # z
    corners[..., 0:4, 2] = aabb[..., 0, 2]
    corners[..., 4:8, 2] = aabb[..., 1, 2]

    corners = corners @ matrix
    out[..., 0, :] = np.min(corners[..., :-1], axis=-2)
    out[..., 1, :] = np.max(corners[..., :-1], axis=-2)

    return out


def quat_to_axis_angle(quaternion, /, *, out=None, dtype=None):
    """Convert a quaternion to axis-angle representation.

    Parameters
    ----------
    quaternion : ndarray, [4]
        A quaternion describing the rotation.
    out : Tuple[ndarray, ...], optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    axis : ndarray, [3]
        The axis around which the quaternion rotates in euclidean coordinates.
    angle : ndarray, [1]
        The angle (in rad) by which the quaternion rotates.

    Notes
    -----
    To use `out` with a single quaternion you need to provide a ndarray of shape
    ``(1,)`` for angle.

    """

    quaternion = np.asarray(quaternion)

    if out is None:
        quaternion = quaternion.astype(dtype)
        out = (
            quaternion[..., :3] / np.sqrt(1 - quaternion[..., 3] ** 2),
            2 * np.arccos(quaternion[..., 3]),
        )
    else:
        out[0][:] = quaternion[..., :3] / np.sqrt(1 - quaternion[..., 3] ** 2)
        out[1][:] = 2 * np.arccos(quaternion[..., 3])

    return out


__all__ = [
    name for name in globals() if name.startswith(("vec_", "mat_", "quat_", "aabb_"))
]



"""Note that we assume unit quaternions for faster implementations"""

import numpy as np
from numpy.lib.stride_tricks import as_strided


def mat_from_quat(quaternion, /, *, out=None, dtype=None):
    """
    Make a rotation matrix given a quaternion.

    Parameters
    ----------
    quaternion : ndarray, [4]
        Quaternion.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4, 4]
        rotation matrix.
    """
    quaternion = np.asarray(quaternion)
    result_shape = (*quaternion.shape[:-1], 4, 4)

    if out is None:
        out = np.empty(result_shape, dtype=dtype)

    # view into the diagonal of the result
    n_matrices = np.prod(result_shape[:-2], dtype=int)
    itemsize = out.itemsize
    diagonal = as_strided(
        out, shape=(n_matrices, 4), strides=(16 * itemsize, 5 * itemsize)
    )

    out[:] = 0
    diagonal[:] = 1

    x, y, z, w = (
        quaternion[..., 0],
        quaternion[..., 1],
        quaternion[..., 2],
        quaternion[..., 3],
    )

    x2 = x * 2
    y2 = y * 2
    z2 = z * 2
    xx = x * x2
    xy = x * y2
    xz = x * z2
    yy = y * y2
    yz = y * z2
    zz = z * z2
    wx = w * x2
    wy = w * y2
    wz = w * z2

    out[..., 0, 0] = 1 - (yy + zz)
    out[..., 1, 0] = xy + wz
    out[..., 2, 0] = xz - wy
    out[..., 0, 1] = xy - wz
    out[..., 1, 1] = 1 - (xx + zz)
    out[..., 2, 1] = yz + wx
    out[..., 0, 2] = xz + wy
    out[..., 1, 2] = yz - wx
    out[..., 2, 2] = 1 - (xx + yy)

    return out


def quat_mul(a, b, /, *, out=None, dtype=None):
    """
    Multiply two quaternions

    Parameters
    ----------
    a : ndarray, [4]
        Left-hand quaternion
    b : ndarray, [4]
        Right-hand quaternion
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4]
        Quaternion.
    """
    a = np.asarray(a)
    b = np.asarray(b)

    if out is None:
        out = np.empty(4, dtype=dtype)

    xyz = a[3] * b[:3] + b[3] * a[:3] + np.cross(a[:3], b[:3])
    w = a[3] * b[3] - a[:3].dot(b[:3])

    out[:3] = xyz
    out[3] = w

    return out


def quat_from_vecs(source, target, /, *, out=None, dtype=None):
    """Rotate one vector onto another.

    Create a quaternion that rotates ``source`` onto ``target``.

    Parameters
    ----------
    source : ndarray, [3]
        The vector that should be rotated.
    target : ndarray, [3]
        The vector that will be rotated onto.
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4]
        Quaternion.

    Notes
    -----
    Among all the possible rotations that send ``source`` onto ``target`` this
    function always chooses the right-hand rotation around the origin. In cases
    where more than one right-hand rotation around the origin exists (``source``
    and ``target`` are parallel), an arbitrary one is returned.

    While this function is intended to be used with unit vectors, it also works
    on non-unit vectors in which case the returned rotation will point
    ``source`` in the direction of ``target``.

    """

    source = np.asarray(source, dtype=float)
    target = np.asarray(target, dtype=float)

    if out is None:
        result_shape = np.broadcast_shapes(source.shape, target.shape)[:-1]
        out = np.empty((*result_shape, 4), dtype=dtype)

    axis = np.cross(source, target)
    angle = np.arctan2(np.linalg.norm(axis), np.dot(source, target))

    # if source and target are parallel, axis will be 0. In this case, we
    # need to choose a replacement axis, which is any vector that is orthogonal
    # to source (and/or target).
    use_fallback = np.linalg.norm(axis, axis=-1) == 0
    if np.any(use_fallback):
        fallback = np.empty((*use_fallback.shape, 3), dtype=float)
        fallback = np.atleast_2d(fallback)

        template = source[use_fallback]
        y_zero = template[..., 1] == 0
        z_zero = template[..., 2] == 0
        both_nonzero = ~(y_zero | z_zero)

        # if any axis is zero, we can use that axis
        fallback[y_zero, :] = (0, 1, 0)
        fallback[z_zero, :] = (0, 0, 1)

        # if two axes are non-zero we can use those
        if np.any(both_nonzero):
            fallback[both_nonzero, :] = (0, -1, 1) * template[both_nonzero, [0, 2, 1]]

        axis[use_fallback] = np.squeeze(fallback)

    return quat_from_axis_angle(axis, angle, out=out)


def quat_inv(quaternion, /, *, out=None, dtype=None):
    """
    Inverse of a given quaternion

    Parameters
    ----------
    a : ndarray, [3]
        First unit vector
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4]
        Quaternion.
    """
    quaternion = np.asarray(quaternion)

    if out is None:
        out = np.empty_like(quaternion, dtype=dtype)

    out[:] = quaternion
    out[..., :3] *= -1

    return out


def quat_from_axis_angle(axis, angle, /, *, out=None, dtype=None):
    """Quaternion from axis-angle pair.

    Create a quaternion representing the rotation of an given angle
    about a given unit vector

    Parameters
    ----------
    axis : ndarray, [3]
        Unit vector
    angle : number
        The angle (in radians) to rotate about axis
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    ndarray, [4]
        Quaternion.
    """

    axis = np.asarray(axis, dtype=float)
    angle = np.asarray(angle, dtype=float)

    if out is None:
        out_shape = np.broadcast_shapes(axis.shape[:-1], angle.shape)
        out = np.empty((*out_shape, 4), dtype=dtype)

    # result should be independent of the length of the given axis
    lengths_shape = axis.shape[:-1] + (1,)
    axis /= np.linalg.norm(axis, axis=-1).reshape(lengths_shape)

    out[..., :3] = axis * np.sin(angle / 2).reshape(lengths_shape)
    out[..., 3] = np.cos(angle / 2)

    return out


def quat_from_euler(angles, /, *, order="xyz", out=None, dtype=None):
    """
    Create a quaternion from euler angles.

    Parameters
    ----------
    angles : ndarray, [3]
        A set of XYZ euler angles.
    order : string, optional
        The rotation order as a string. Can include 'X', 'Y', 'Z' for intrinsic
        rotation (uppercase) or 'x', 'y', 'z' for extrinsic rotation (lowercase).
        Default is "xyz".
    out : ndarray, optional
        A location into which the result is stored. If provided, it
        must have a shape that the inputs broadcast to. If not provided or
        None, a freshly-allocated array is returned. A tuple must have
        length equal to the number of outputs.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    quaternion : ndarray, [4]
        The rotation expressed as a quaternion.

    """

    angles = np.asarray(angles, dtype=float)
    batch_shape = angles.shape[:-1] if len(order) > 1 else angles.shape

    if out is None:
        out = np.empty((*batch_shape, 4), dtype=dtype)

    # work out the sequence in which to apply the rotations
    is_extrinsic = [x.islower() for x in order]
    basis_index = {"x": 0, "y": 1, "z": 2}
    order = [basis_index[x] for x in order.lower()]

    # convert each euler matrix into a quaternion
    quaternions = np.zeros((len(order), *batch_shape, 4), dtype=float)
    quaternions[:, ..., -1] = np.cos(angles / 2)
    quaternions[np.arange(len(order)), ..., order] = np.sin(angles / 2)

    # multiple euler-angle quaternions respecting
    out[:] = quaternions[0]
    for idx in range(1, len(quaternions)):
        if is_extrinsic[idx]:
            quat_mul(quaternions[idx], out, out=out)
        else:
            quat_mul(out, quaternions[idx], out=out)

    return out


__all__ = [
    name for name in globals() if name.startswith(("vec_", "mat_", "quat_", "aabb_"))
]
